from .insn import *

"""
RV32I S-Type Instruction
"""

class InsnRV32ISType(Insn):
    def __init__(self, params, funct3, mask_shift):
        super().__init__(params)
        self.funct3 = funct3
        self.mask_shift = mask_shift
        self.addr = Signal(self.params.xlen)

    def elaborate(self, platform):
        m = super().elaborate(platform)

        m.d.comb += self.insn_imm.eq(Value.as_signed(Cat(self.rvfi_insn[7:12], self.rvfi_insn[25:32])))
        
        if self.params.csr_misa:
            m.d.comb += self.misa_ok.eq((self.rvfi_csr_misa_rdata & 0) == 0)
            m.d.comb += self.spec_csr_misa_rmask.eq(0)
        else:
            m.d.comb += self.misa_ok.eq(1)

        if self.params.aligned_mem:
            m.d.comb += self.addr.eq(self.rvfi_rs1_rdata + self.insn_imm)
            m.d.comb += self.spec_valid.eq(self.rvfi_valid & (~self.insn_padding) & (self.insn_funct3 == self.funct3) & (self.insn_opcode == 0b0100011))
            m.d.comb += self.spec_rs1_addr.eq(self.insn_rs1)
            m.d.comb += self.spec_rs2_addr.eq(self.insn_rs2)
            m.d.comb += self.spec_mem_addr.eq(self.addr & ~(int(self.params.xlen // 8) - 1))
            m.d.comb += self.spec_mem_wmask.eq(((1 << self.mask_shift) - 1) << (self.addr - self.spec_mem_addr)[:5])
            m.d.comb += self.spec_mem_wdata.eq(self.rvfi_rs2_rdata << (8 * (self.addr - self.spec_mem_addr))[:8])
            m.d.comb += self.spec_pc_wdata.eq(self.rvfi_pc_rdata + 4)
            m.d.comb += self.spec_trap.eq(((self.addr & (self.mask_shift - 1)) != 0) | ~self.misa_ok)
        else:
            m.d.comb += self.addr.eq(self.rvfi_rs1_rdata + self.insn_imm)
            m.d.comb += self.spec_valid.eq(self.rvfi_valid & (~self.insn_padding) & (self.insn_funct3 == self.funct3) & (self.insn_opcode == 0b0100011))
            m.d.comb += self.spec_rs1_addr.eq(self.insn_rs1)
            m.d.comb += self.spec_rs2_addr.eq(self.insn_rs2)
            m.d.comb += self.spec_mem_addr.eq(self.addr)
            m.d.comb += self.spec_mem_wmask.eq((1 << self.mask_shift) - 1)
            m.d.comb += self.spec_mem_wdata.eq(self.rvfi_rs2_rdata)
            m.d.comb += self.spec_pc_wdata.eq(self.rvfi_pc_rdata + 4)
            m.d.comb += self.spec_trap.eq(~self.misa_ok)

        return m
